==> dirac_0.5.2.orig.tar.gz: dirac-0.5.2/libdirac_encoder/quality_monitor.cpp <==
/* ***** BEGIN LICENSE BLOCK *****
*
* $Id: quality_monitor.cpp,v 1.12 2005/02/02 14:47:18 tjdwave Exp $ $Name: Dirac_0_5_2 $
*
* Version: MPL 1.1/GPL 2.0/LGPL 2.1
*
* The contents of this file are subject to the Mozilla Public License
* Version 1.1 (the "License"); you may not use this file except in compliance
* with the License. You may obtain a copy of the License at
* http://www.mozilla.org/MPL/
*
* Software distributed under the License is distributed on an "AS IS" basis,
* WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
* the specific language governing rights and limitations under the License.
*
* The Original Code is BBC Research and Development code.
*
* The Initial Developer of the Original Code is the British Broadcasting
* Corporation.
* Portions created by the Initial Developer are Copyright (C) 2004.
* All Rights Reserved.
*
* Contributor(s): Thomas Davies (Original Author)
*
* Alternatively, the contents of this file may be used under the terms of
* the GNU General Public License Version 2 (the "GPL"), or the GNU Lesser
* Public License Version 2.1 (the "LGPL"), in which case the provisions of
* the GPL or the LGPL are applicable instead of those above. If you wish to
* allow use of your version of this file only under the terms of the either
* the GPL or LGPL and not to allow others to use your version of this file
* under the MPL, indicate your decision by deleting the provisions above
* and replace them with the notice and other provisions required by the GPL
* or LGPL. If you do not delete the provisions above, a recipient may use
* your version of this file under the terms of any one of the MPL, the GPL
* or the LGPL.
* ***** END LICENSE BLOCK ***** */

#include <libdirac_encoder/quality_monitor.h>
#include <libdirac_common/wavelet_utils.h>
using namespace dirac;

using std::log10;

QualityMonitor::QualityMonitor(EncoderParams& encp, 
                               const SeqParams& sparams)
:
    m_encparams(encp),
    m_cformat( sparams.CFormat() ),
    m_true_xl( sparams.Xl() ),
    m_true_yl( sparams.Yl() ),
    m_target_quality(3),
    m_last_quality(3),
    m_slope(3),
    m_offset(3),
    m_last_lambda(3)
{
    ResetAll();
}

void QualityMonitor::ResetAll()
{
    // set target qualities
 	m_target_quality[I_frame] = 0.28 * m_encparams.Qf()* m_encparams.Qf() + 20.0 ;
	m_target_quality[L1_frame] = m_target_quality[I_frame] - 1.0;
	m_target_quality[L2_frame] = m_target_quality[I_frame] - 2.0;

    // assume we hit those targets last time
    m_last_quality = m_target_quality;

    // set defaults for the model
     m_slope[I_frame] = -4.0;
     m_slope[L1_frame] = -4.0;
     m_slope[L2_frame] = -4.0;
     m_offset[I_frame] = 38.5,
     m_offset[L1_frame] = 43.3;
     m_offset[L2_frame] = 43.3;

    for (size_t fsort=0; fsort<3; ++fsort)
    {
        m_last_lambda[fsort] = std::pow( 10.0, (m_target_quality[fsort] - m_offset[fsort])/m_slope[fsort] );
    }// fsort

    // set a default ratio for the motion estimation lambda
    // Exact value TBD - will incorporate stuff about blocks and so on
    // Also need to think about how this can be adapted for sequences with more or less motion 

    m_me_ratio = 0.1;

 	// set up the Lagrangian parameters
    for (size_t fsort=0; fsort<3; ++fsort)
    {
        m_encparams.SetLambda( FrameSort(fsort), m_last_lambda[fsort] );
    }// fsort

    m_encparams.SetL1MELambda( std::sqrt(m_encparams.L1Lambda())*m_me_ratio );
    m_encparams.SetL2MELambda( std::sqrt(m_encparams.L2Lambda())*m_me_ratio );

}

bool QualityMonitor::UpdateModel(const Frame& ld_frame, const Frame& orig_frame , const int count)
{
    // The return value - true if we need to recode, false otherwise
    bool recode = false;

	const FrameSort& fsort = ld_frame.GetFparams().FSort();	
	double target_quality;	

	// Parameters relating to the last frame we measured
	double last_lambda;
	double last_quality;

	// Parameters relating to the current frame
	double current_lambda;
	double current_quality;


    // Set up local parameters for the particular frame type
    current_lambda = m_encparams.Lambda(fsort);
    last_lambda = m_last_lambda[fsort];
    last_quality = m_last_quality[fsort];
    target_quality = m_target_quality[fsort];

    // Get the quality of the current frame
	current_quality = QualityVal( ld_frame.Ydata() , orig_frame.Ydata() , 0.0 , fsort );

    // Copy current data into memory for last frame data
    m_last_lambda[fsort] = m_encparams.Lambda(fsort);
    m_last_quality[fsort] = current_quality;

	// Ok, so we've got an actual quality value to use. We know the lambda used before and the resulting
	// quality then allows us to estimate the slope of the curve of quality versus log of lambda

	if ( std::abs(current_quality - last_quality)> 0.2 && 
         std::abs(log10(current_lambda) - log10(last_lambda)) > 0.1 ) 
	{// if we can adjust model accurately, do so

        double slope, offset;

        // Calculate the slope of WPSNR versus log(lambda) from prior measurements
 	    slope = (current_quality - last_quality)/( log10(current_lambda) - log10(last_lambda) );
 
        //Restrict so that the value isn't too extreme
        slope = std::min( std::max( -10.0 , slope ), -0.1);

  		// Calculate the resulting offset
		offset = current_quality - ( log10(current_lambda) * slope );

        if ( count != 1 )
        {
            // Update the default values using a simple recursive filter ...
            m_slope[fsort] = (3.0*m_slope[fsort] + slope)/4.0;
            m_offset[fsort] = (3.0*m_offset[fsort] + offset)/4.0;            
        }
        else
        {
            // .. unless we're recoding a frame for the first time            
            m_slope[fsort] = (m_slope[fsort] + slope)/2.0;
            m_offset[fsort] = (m_offset[fsort] + offset)/2.0;
        }
        m_slope[fsort] = std::min( std::max( -10.0 , m_slope[fsort] ), -1.5);
    }

    // If we need to adjust the lambdas, do so
	if ( std::abs(current_quality - target_quality)> 0.2 )
	{
        // Update the lambdas as appropriate
        float quality_diff = m_target_quality[fsort] - current_quality;

        CalcNewLambdas(fsort , std::min( m_slope[fsort] , -1.0 ), quality_diff );
    }

    // if we have a large difference in quality, recode)
    if ( std::abs( current_quality - target_quality )>1.5 )
        recode = true;

    return recode;
}

void QualityMonitor::CalcNewLambdas(const FrameSort fsort, const double slope, const double quality_diff )
{	
     const double clipped_quality_ratio = std::min( 2.0 , std::max( quality_diff/slope , -2.0 ) );

     if ( m_encparams.Lambda(fsort) > 100001.0 && clipped_quality_ratio > 0.0 )
         m_encparams.SetLambda(fsort, 100000.0);
     else
         m_encparams.SetLambda(fsort, m_encparams.Lambda(fsort) *
                             std::pow( (double)10.0, clipped_quality_ratio ) );

     if (fsort == L1_frame)
 		m_encparams.SetL1MELambda( std::sqrt(m_encparams.L1Lambda()) * m_me_ratio );
     else if (fsort == L2_frame)
 		m_encparams.SetL2MELambda( std::sqrt(m_encparams.L2Lambda()) * m_me_ratio );

}

double QualityMonitor::QualityVal(const PicArray& coded_data, const PicArray& orig_data , double cpd , const FrameSort fsort)
{

    // The number of regions to look at in assessing quality
    int xregions( 4 );
    int yregions( 3 );

    if ( fsort == I_frame )
    {
        xregions = 1;
        yregions = 1;
    }

    TwoDArray<long double> diff_array( yregions , xregions);
	long double diff;
    ValueType filt_orig, filt_coded;

    OneDArray<int> xstart( diff_array.LengthX() );
    OneDArray<int> xend( diff_array.LengthX() );
    OneDArray<int> ystart( diff_array.LengthY() );
    OneDArray<int> yend( diff_array.LengthX() );

    for ( int i=0 ; i<xstart.Length() ; ++i)
    { 
        xstart[i] =( i * (m_true_xl-2) )/xstart.Length()+1;
        xend[i] = ( (i+1) * (m_true_xl-2) )/xstart.Length()+1;
    }

    for ( int i=0 ; i<ystart.Length() ; ++i)
    { 
        ystart[i] =( i * (m_true_yl-2) )/ystart.Length()+1;
        yend[i] = ( (i+1) * (m_true_yl-2) )/ystart.Length()+1;
    }

    for ( int q=0 ; q<diff_array.LengthY() ; q++ )
    { 
        for ( int p=0 ; p<diff_array.LengthX() ; p++ )
        { 
            diff_array[q][p] = 0.0;

            for (int j=ystart[q]; j<yend[q]; j++)
            {
                for (int i=xstart[p]; i<xend[p]; i++)
                {
                    filt_coded = Filter( coded_data , i , j );
                    filt_orig = Filter( orig_data , i , j );

                    diff = static_cast<long double> ( filt_coded - filt_orig );
 
                    diff *= diff;
                    diff *= diff;

                    diff_array[q][p] += diff;
                }//i
            }//j

            diff_array[q][p] /= ( xend[p]-xstart[p] ) * ( yend[q]-ystart[q] );
            diff_array[q][p] = std::sqrt( diff_array[q][p] );

            // now compensate for the fact that we've got two extra bits
            diff_array[q][p] /= 16.0;

        }// p
    }// q
     
    // return the worst area
    long double worst_diff = diff_array[0][0];
    for ( int q=0 ; q<diff_array.LengthY() ; ++q )
    { 
        for ( int p=0 ; p<diff_array.LengthX() ; ++p )
        { 
            if ( diff_array[q][p] > worst_diff )          
                worst_diff = diff_array[q][p];
        }// p
    }// q

	return static_cast<double> ( 10.0 * std::log10( 255.0*255.0 / worst_diff ) );	
}

ValueType QualityMonitor::Filter( const PicArray& data , const int xpos , const int ypos ) const
{
    return (
            int(data[ypos-1][xpos-1]) +6*int(data[ypos-1][xpos]) + int(data[ypos-1][xpos+1]) + 
            6*int(data[ypos][xpos-1]) + 36*int(data[ypos][xpos]) + 6*int(data[ypos][xpos+1]) + 
            int(data[ypos+1][xpos-1]) +6*int(data[ypos+1][xpos]) + int(data[ypos+1][xpos+1]) +
            32 )>>6;
}

==> dirac_0.5.4.orig.tar.gz: dirac-0.5.4/libdirac_encoder/quality_monitor.cpp <==
/* ***** BEGIN LICENSE BLOCK *****
*
* $Id: quality_monitor.cpp,v 1.16 2005/09/27 13:08:05 tjdwave Exp $ $Name: Dirac_0_5_4 $
*
* Version: MPL 1.1/GPL 2.0/LGPL 2.1
*
* The contents of this file are subject to the Mozilla Public License
* Version 1.1 (the "License"); you may not use this file except in compliance
* with the License. You may obtain a copy of the License at
* http://www.mozilla.org/MPL/
*
* Software distributed under the License is distributed on an "AS IS" basis,
* WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
* the specific language governing rights and limitations under the License.
*
* The Original Code is BBC Research and Development code.
*
* The Initial Developer of the Original Code is the British Broadcasting
* Corporation.
* Portions created by the Initial Developer are Copyright (C) 2004.
* All Rights Reserved.
*
* Contributor(s): Thomas Davies (Original Author)
*
* Alternatively, the contents of this file may be used under the terms of
* the GNU General Public License Version 2 (the "GPL"), or the GNU Lesser
* Public License Version 2.1 (the "LGPL"), in which case the provisions of
* the GPL or the LGPL are applicable instead of those above. If you wish to
* allow use of your version of this file only under the terms of the either
* the GPL or LGPL and not to allow others to use your version of this file
* under the MPL, indicate your decision by deleting the provisions above
* and replace them with the notice and other provisions required by the GPL
* or LGPL. If you do not delete the provisions above, a recipient may use
* your version of this file under the terms of any one of the MPL, the GPL
* or the LGPL.
* ***** END LICENSE BLOCK ***** */

#include <libdirac_encoder/quality_monitor.h>
#include <libdirac_common/wavelet_utils.h>
using namespace dirac;

using std::log10;

QualityMonitor::QualityMonitor(EncoderParams& encp, 
                               const SeqParams& sparams)
:
    m_encparams(encp),
    m_cformat( sparams.CFormat() ),
    m_true_xl( sparams.Xl() ),
    m_true_yl( sparams.Yl() ),
    m_quality_average(3),
    m_frame_total(3)
{
    ResetAll();
}

QualityMonitor::~QualityMonitor()
{}

void QualityMonitor::ResetAll()
{
    // Set the lambdas

    if ( !m_encparams.Lossless() )
    {
        m_encparams.SetLambda( I_frame , std::pow( 10.0 , (10.0-m_encparams.Qf() )/2.5 ) );
        m_encparams.SetLambda( L1_frame , m_encparams.ILambda()*128.0 );
        m_encparams.SetLambda( L2_frame , m_encparams.ILambda()*512.0 );


        // Set the lambdas for motion estimation
        const double me_ratio = 2.0;

        m_encparams.SetL1MELambda( std::sqrt(m_encparams.L1Lambda())*me_ratio );
        m_encparams.SetL2MELambda( std::sqrt(m_encparams.L2Lambda())*me_ratio );

        for (int i=0; i<3 ; ++i )
        {
            m_quality_average[i] = 0.0;
            m_frame_total[i] = 0;
        }// i
    }
    else
    {
        m_encparams.SetLambda( I_frame , 0.0 );
        m_encparams.SetLambda( L1_frame , 0.0 );
        m_encparams.SetLambda( L2_frame , 0.0 );

        m_encparams.SetL1MELambda( 0.0 );
        m_encparams.SetL2MELambda( 0.0 );
    }
}

void QualityMonitor::WriteLog()
{
    std::cerr<<std::endl<<"Mean quality for I frames is "<<m_quality_average[I_frame]/m_frame_total[I_frame];
    std::cerr<<std::endl<<"Mean quality for L1 frames is "<<m_quality_average[L1_frame]/m_frame_total[L1_frame];
    std::cerr<<std::endl<<"Mean quality for L2 frames is "<<m_quality_average[L2_frame]/m_frame_total[L2_frame]<<std::endl;
}

void QualityMonitor::UpdateModel(const Frame& ld_frame, const Frame& orig_frame )
{
	const FrameSort& fsort = ld_frame.GetFparams().FSort();	

	m_quality_average[fsort] += QualityVal( ld_frame.Ydata() , orig_frame.Ydata() , 0.0 , fsort );
    m_frame_total[fsort]++;

}


double QualityMonitor::QualityVal(const PicArray& coded_data, const PicArray& orig_data , double cpd , const FrameSort fsort)
{

    // The number of regions to look at in assessing quality
    const int xregions( 4 );
    const int yregions( 3 );

    TwoDArray<long double> diff_array( yregions , xregions);
    TwoDArray<long double> diff_sq_array( yregions , xregions);
	long double diff;
    ValueType filt_orig, filt_coded;

    OneDArray<int> xstart( diff_array.LengthX() );
    OneDArray<int> xend( diff_array.LengthX() );
    OneDArray<int> ystart( diff_array.LengthY() );
    OneDArray<int> yend( diff_array.LengthX() );

    for ( int i=0 ; i<xstart.Length() ; ++i)
    { 
        xstart[i] =( i * (m_true_xl-2) )/xstart.Length()+1;
        xend[i] = ( (i+1) * (m_true_xl-2) )/xstart.Length()+1;
    }

    for ( int i=0 ; i<ystart.Length() ; ++i)
    { 
        ystart[i] =( i * (m_true_yl-2) )/ystart.Length()+1;
        yend[i] = ( (i+1) * (m_true_yl-2) )/ystart.Length()+1;
    }

    for ( int q=0 ; q<diff_array.LengthY() ; q++ )
    { 
        for ( int p=0 ; p<diff_array.LengthX() ; p++ )
        { 
            diff_sq_array[q][p] = 0.0;

            for (int j=ystart[q]; j<yend[q]; j++)
            {
                for (int i=xstart[p]; i<xend[p]; i++)
                {
                    filt_coded = Filter( coded_data , i , j );
                    filt_orig = Filter( orig_data , i , j );

                    diff = static_cast<long double> ( filt_coded - filt_orig );
 
                    diff *= diff;
                    diff *= diff;

                    diff_sq_array[q][p] += diff;
                }//i
            }//j

            diff_sq_array[q][p] /= ( xend[p]-xstart[p] ) * ( yend[q]-ystart[q] );

            // now compensate for the fact that we've got two extra bits
            diff_sq_array[q][p] /= 256.0;

            diff_array[q][p] = std::sqrt( diff_sq_array[q][p] );

        }// p
    }// q
     
    // return the self-weighted average

    long double sum_diff( 0 );
    long double sum_sq_diff( 0 );
    for ( int q=0 ; q<diff_array.LengthY() ; ++q )
    { 
        for ( int p=0 ; p<diff_array.LengthX() ; ++p )
        { 
            sum_diff += diff_array[q][p];
            sum_sq_diff += diff_sq_array[q][p];
        }// p
    }// q

	return static_cast<double> ( 10.0 * std::log10( 255.0*255.0*sum_diff / sum_sq_diff ) );	
}

ValueType QualityMonitor::Filter( const PicArray& data , const int xpos , const int ypos ) const
{
    return (
            int(data[ypos-1][xpos-1]) +6*int(data[ypos-1][xpos]) + int(data[ypos-1][xpos+1]) + 
            6*int(data[ypos][xpos-1]) + 36*int(data[ypos][xpos]) + 6*int(data[ypos][xpos+1]) + 
            int(data[ypos+1][xpos-1]) +6*int(data[ypos+1][xpos]) + int(data[ypos+1][xpos+1]) +
            32 )>>6;
}
